from .elements import AbstractElement
from .integer import Integer
from ..parser import Parser
from ..utils import fix_return_string, rec_elements


class Integers(AbstractElement):
    def __init__(self, total):
        self._total = total
        self._tokens = [Integer() for _ in range(total)]

    def parse(self, testcase: Parser):
        for token in self._tokens:
            if not token.parse(testcase):
                return False
        return True

    @fix_return_string
    def generate(self):
        variables = ', '.join(tok.name for tok in self._tokens)
        gen = ', '.join('randint({min}, {max})'.format(min=tok._min, max=tok._max) for tok in self._tokens)

        return """
        {variables} = {gen}
        print({variables})
        """.format(variables=variables, gen=gen)

    @rec_elements
    def elements(self):
        return self._tokens

    def __getitem__(self, idx):
        return self._tokens[idx]



import random
from os import listdir
from os.path import join

import bs4

CODEFORCES_PROBLEMS = '.'


def parse_testcase(name):
    with open(join(CODEFORCES_PROBLEMS, name)) as f:
        html = f.read()

    soup = bs4.BeautifulSoup(html, features="html5lib")
    sample_test = soup.find("div", {"class": "sample-tests"})

    tc = []

    for node_inp in sample_test.find_all("div", {"class": "input"}):
        content = node_inp.find("pre")
        content = str(content)
        content = content[5:-6]
        content = content.replace('<br/>', '\n')
        tc.append(content)

    return tc


def codeforces_problems():
    return listdir(CODEFORCES_PROBLEMS)


def random_problem():
    probs = codeforces_problems()
    prob = random.choice(probs)
    tc = parse_testcase(prob)
    return prob, tc


def get_problem(name=None):
    if name is None:
        prob, tc = random_problem()
    else:
        tc = parse_testcase(name)
        prob = name
    return prob, tc


if __name__ == '__main__':
    print(random_problem())



from functools import partial

from ._pattern import PATTERNS
from .language import code

VERBOSE = False


def register(builder):
    """
    Convert pattern builder into compiled pattern.
    Keep __name__.
    """

    if VERBOSE:
        print("Register:", builder.__name__)

    def lazy():
        pattern = code()

        for element in builder():
            pattern.add(element)

        prog = pattern.compile()
        prog.__name__ = builder.__name__
        return prog

    PATTERNS.append(lazy)


def dict_get(dic, key, default):
    if key in dic:
        return dic[key]
    else:
        return default()


def get_interval(interval_str, name_dict):
    # Try to parse as int
    if '..' not in interval_str:
        lower = int(interval_str)
        upper = lower + 1
        return lower, upper

    lower, upper = interval_str.split('..')
    add = upper.startswith('=')
    if add: upper = upper[1:]

    lower = dict_get(name_dict, lower, partial(int, lower))
    upper = dict_get(name_dict, upper, partial(int, upper))

    if add: upper += 1

    return lower, upper


def generate(ordered_params, params_idx, name_dict):
    if params_idx == len(ordered_params):
        yield name_dict

    else:
        name, value = ordered_params[params_idx]
        lower, upper = get_interval(value, name_dict)

        for v in range(lower, upper):
            name_dict[name] = v

            for param_variants in generate(ordered_params, params_idx + 1, name_dict):
                yield param_variants

        if name in name_dict:
            del name_dict[name]


def generate_all(ordered_params):
    name_dict = {}
    return generate(ordered_params, 0, name_dict)


def register_all(**params):
    ordered_params = list(params.items())

    def proc(func):
        for param_variants in generate_all(ordered_params):
            new_func = partial(func, **param_variants)
            new_func.__name__ = "{}({})".format(
                func.__name__,
                ','.join(str(v) for k, v in param_variants.items())
            )
            register(new_func)

    return proc



class Parser:
    BLANK = ' \n\t'

    def __init__(self, text):
        self._text = text
        self._pointer = 0

    def peek(self):
        if self._pointer == len(self._text):
            return None
        else:
            return self._text[self._pointer]

    def get(self):
        c = self.peek()
        if c is not None:
            self._pointer += 1
        return c

    @staticmethod
    def is_readable(ch):
        return ch is not None and ch not in Parser.BLANK

    def token(self, allow_blank=True):
        if allow_blank:
            while self.peek() is not None and self.peek() in Parser.BLANK:
                self.get()

        if not self.is_readable(self.peek()):
            raise ParserError("Token expected.")

        token = ""

        while True:
            if not self.is_readable(self.peek()):
                break
            token += self.get()

        return token

    def consume_blank(self):
        while True:
            x = self.peek()
            if x is not None and x in Parser.BLANK:
                self.get()
            else:
                break


class ParserError(BaseException):
    pass



from .elements import AbstractElement
from .parser import Parser
from .utils import names, element_iterator


class Code:
    def __init__(self):
        self._elements = []

    def add(self, elem: AbstractElement):
        self._elements.append(elem)
        return elem

    def assign_names(self):
        # assign names to every element that needs it
        used_names = set()

        for elem in element_iterator(self._elements):
            if hasattr(elem, 'name'):
                if elem.name is not None:
                    if elem.name in used_names:
                        # Name was previously used
                        elem.name = None
                    else:
                        used_names.add(elem.name)

        for elem in element_iterator(self._elements):
            if hasattr(elem, 'name'):
                if elem.name is None:
                    for pref_name in names(elem.PREFERRED_NAMES):
                        if pref_name not in used_names:
                            used_names.add(pref_name)
                            elem.name = pref_name
                            break

    def compile(self):
        self.assign_names()
        return Program(self._elements)


class Program:
    def __init__(self, elements):
        self._elements = elements
        self._testcases = None
        self._ok = True

    @property
    def ok(self):
        return self._ok

    def _parse(self, tc):
        tc = Parser(tc)

        for elem in self._elements:
            self._ok = elem.parse(tc)

            # Break parsing since it failed parsing object `elem`
            if not self._ok:
                return

        tc.consume_blank()
        self._ok = tc.peek() is None

    def parse(self, testcases):
        self._testcases = testcases

        for tc in testcases:
            self._parse(tc)

            if not self.ok:
                return False

        return True

    @property
    def header(self):
        return "from random import randint, choice\n"

    def generate(self):

        body = '\n\n'.join(
            obj.generate() for obj in self._elements
        ) + '\n'

        code = self.header + '\n' + body
        return code


code = Code



from ..parser import Parser

PREFERRED_NAMES = ['n', 'm', 'k', 'a', 'b', 'c', 'd', 't', 'q']


class AbstractElement:

    def parse(self, testcase: Parser):
        raise NotImplementedError()

    def generate(self):
        raise NotImplementedError()

    def elements(self):
        raise NotImplementedError()



from tqdm import tqdm, TqdmLogger as logger

from codeforces.tools import codeforces_problems, get_problem
from generate import find_pattern


class Tabular:
    def __init__(self):
        self._dic = {}

    def add(self, name):
        self._dic[name] = self._dic.get(name, 0) + 1

    def report(self):
        L = [(v, k) for k, v in self._dic.items()]
        L.sort(reverse=True)
        print()
        for v, k in L:
            print("{:<8} {}".format(v, k))


def explore():
    total = 0
    ok = 0

    tab = Tabular()

    for idx, prob in tqdm(enumerate(codeforces_problems())):
        try:
            p, tc = get_problem(prob)
            pat = find_pattern(tc)

            total += 1

            if pat is not None:
                tab.add(pat.__name__)
                ok += 1
                logger.log("{}/{}".format(ok, total))
        except:
            print("Fail on problem: {}".format(prob))

        if (idx + 1) % 100 == 0:
            tab.report()

    tab.report()


if __name__ == '__main__':
    # print(codeforces_problems()[533])
    explore()



from .elements import AbstractElement, PREFERRED_NAMES
from ..parser import Parser, ParserError
from ..utils import fix_return_string, rec_elements


def get_value(obj):
    if isinstance(obj, int):
        return obj
    elif isinstance(obj, Integer):
        return obj._last_value
    elif isinstance(obj, Operation):
        return obj.value


class Integer(AbstractElement):
    PREFERRED_NAMES = PREFERRED_NAMES

    def __init__(self):
        self._min = float('inf')
        self._max = float('-inf')
        self._last_value = None
        self.name = None

    def parse(self, testcase: Parser):
        # Read next token
        try:
            value = testcase.token()
        except ParserError:
            return False

        # Convert into integer
        try:
            value = int(value)
        except ValueError:
            return False

        self._min = min(self._min, value)
        self._max = max(self._max, value)
        self._last_value = value
        return True

    @fix_return_string
    def generate(self):
        return """
        {name} = randint({min}, {max})
        print({name})
        """.format(name=self.name, min=self._min, max=self._max)

    @rec_elements
    def elements(self):
        return []


class Operation:
    def __init__(self, left, op, right):
        self.left = left
        self.op = op
        self.right = right

    @property
    def value(self):
        lvalue = get_value(self.left)
        rvalue = get_value(self.right)



from .elements import AbstractElement, PREFERRED_NAMES
from ..parser import Parser, ParserError
from ..utils import fix_return_string


class String(AbstractElement):
    PREFERRED_NAMES = ['s', 'x', 'y', 'z'] + PREFERRED_NAMES

    def __init__(self):
        self._alpha = set()
        self._min_len = float('inf')
        self._max_len = float('-inf')
        self.name = None

    def parse(self, testcase: Parser):
        try:
            value = testcase.token()
        except ParserError:
            return False

        for x in value:
            self._alpha.add(x)

        self._min_len = min(self._min_len, len(value))
        self._max_len = max(self._max_len, len(value))
        return True

    @fix_return_string
    def generate(self):
        return """
        {name} = ''.join([choice("{alpha}") for _ in range(randint({min}, {max}))])
        print({name})
        """.format(name=self.name, alpha=''.join(sorted(self._alpha)), min=self._min_len, max=self._max_len)

    def elements(self):
        return []



from .repository import PATTERNS

DEFAULT = """from random import randint

a = randint(0, 100)
print(a)
"""


def find_pattern(testcase):
    for prog_builder in PATTERNS:
        prog = prog_builder()
        if prog.parse(testcase):
            return prog
    return None


# Simplest strategy.
# Find first pattern that match.
def generate(testcase):
    prog = find_pattern(testcase)

    if prog is None:
        return DEFAULT
    else:
        return prog.generate()



from codeforces.correct_test import test_problems
from codeforces.tools import get_problem
from generate import find_pattern


def run():
    problem, tc = get_problem("839E.html")

    print("name:", problem)
    for xtc in tc:
        print(xtc)

    gen = find_pattern(tc)

    if gen is None:
        print("Pattern not found!")

    else:
        print("pattern:", gen.__name__)
        print(gen.generate())


if __name__ == '__main__':
    assert test_problems()
    run()



from .elements import AbstractElement
from .integer import Integer, Operation, get_value
from ..parser import Parser
from ..utils import fix_return_string, fix_string, indent, rec_elements


class List(AbstractElement):
    PREFERRED_NAMES = ['l'] + Integer.PREFERRED_NAMES

    def __init__(self, total, elem=None):
        if elem is None:
            elem = Integer()

        assert isinstance(total, int) or isinstance(total, Integer) or isinstance(total, Operation)

        self._total = total
        self._token = elem
        self.name = None

    def _upper_limit(self):
        return get_value(self._total)

    def _gen_total(self):
        if isinstance(self._total, int):
            return self._total
        else:
            return self._total.name

    def parse(self, testcase: Parser):
        for i in range(self._upper_limit()):
            if not self._token.parse(testcase):
                return False
        return True

    @fix_return_string
    def generate(self):
        code = fix_string("""
        for _ in range({total}):
        """.format(total=self._gen_total())) + '\n' + indent(self._token.generate(), +1)
        return code

    @rec_elements
    def elements(self):
        yield self._token



from random import randint


def indent(code, size):
    final_eol = code.endswith('\n')
    if final_eol: code = code[:-1]
    lines = code.split('\n')
    lines = [' ' * (4 * size) + line for line in lines]
    code = '\n'.join(lines)
    if final_eol: code += '\n'
    return code


def fix_string(code):
    """
    Fix indentation and empty lines from string coming out of the function.
    """
    lines = code.split('\n')

    # Remove first line if empty
    if not lines[0]:
        lines = lines[1:]

    # Detect tabsize
    size = 0
    while lines[0][size] in ' ':
        size += 1

    lines = list(map(lambda l: l[size:], lines))

    if not lines[-1]:
        lines = lines[:-1]

    code = '\n'.join(lines)
    return code


def fix_return_string(func):
    """
    Decorator that fix the string resulting from a function.
    """

    def _func(*args):
        return fix_string(func(*args))

    return _func


def names(preferred_names):
    """
    Generator of names to feed variables
    """

    def random_word(length):
        return ''.join(map(lambda x: chr(x), [randint(97, 122) for _ in range(length)]))

    for x in preferred_names:
        yield x

    while True:
        name = random_word(3)
        yield name


def rec_elements(elem_iterator):
    def _func(*args):
        for elem in elem_iterator(*args):
            yield elem

            for sub_elem in elem.elements():
                yield sub_elem

    return _func


@rec_elements
def element_iterator(elements):
    return elements



import json
import unittest

from codeforces.tools import get_problem
from generate import find_pattern


def check(prob, pattern_name, with_assert=None):
    _, tc = get_problem(prob)
    gen = find_pattern(tc)

    if with_assert is not None:
        with_assert.assertIsNotNone(gen, "problem {} expected {}".format(prob, pattern_name))
        with_assert.assertEqual(gen.__name__, pattern_name)
        return True
    else:
        if gen is None: return False
        return gen.__name__ == pattern_name


def test_problems(with_assert=None):
    with open('ok.json') as f:
        data = json.load(f)

    for prob, pattern_name in data.items():
        if not check(prob, pattern_name, with_assert):
            return False

    return True


class TestCodeforces(unittest.TestCase):
    def test_problems(self):
        test_problems(self)


if __name__ == '__main__':
    unittest.main()



import unittest

from elements import Integer, List, Integers, Compound, String
from language import code
from parser import Parser, ParserError


def example(test_example):
    def extended_test_example(self):
        pattern = code()
        config = test_example(self, pattern)
        prog = pattern.compile()
        prog.parse(config.get("testcases"))

        if not config.get("should_fail", False):
            self.assertTrue(prog.ok)
            self.assertEqual(prog.generate(), config.get("expected"))
        else:
            self.assertFalse(prog.ok)

    return extended_test_example


class TestLanguageBehavior(unittest.TestCase):
    @example
    def test_one_integer(self, pattern):
        pattern.add(Integer())

        return dict(
            testcases=["1\n", "2\n", "100\n"],
            expected="from random import randint, choice\n\nn = randint(1, 100)\nprint(n)\n",
        )

    @example
    def test_two_integers(self, pattern):
        pattern.add(Integer())
        pattern.add(Integer())

        return dict(
            testcases=["1 3\n", "2 1\n", "100\n2\n"],
            expected='from random import randint, choice\n\nn = randint(1, 100)\nprint(n)\n\n'
                     'm = randint(1, 3)\nprint(m)\n'
        )

    @example
    def test_fixed_two(self, pattern):
        pattern.add(Integers(2))

        return dict(
            testcases=["1 4\n", "2 2\n", "100 -2\n"],
            expected='from random import randint, choice\n\nn, m = randint(1, 100), randint(-2, 4)\nprint(n, m)\n'
        )

    @example
    def test_fail_one_integer(self, pattern):
        pattern.add(Integer())

        return dict(
            testcases=["1 3\n", "2 1\n", "100 2\n"],
            should_fail=True,
        )

    @example
    def test_list(self, pattern):
        f = Integers(2)
        pattern.add(List(3, f))

        return dict(
            testcases=["1 4\n2 2\n3 3\n", "2 2\n1 5\n-1 2\n"],
            expected='from random import randint, choice\n\nfor _ in range(3):'
                     '\n    n, m = randint(-1, 3), randint(2, 5)\n    print(n, m)\n'
        )

    @example
    def test_list_parametric(self, pattern):
        a = pattern.add(Integer())
        pattern.add(List(a, Integers(2)))

        return dict(
            testcases=['\n1\n2 3\n', '3\n1 2\n2 3\n4 5\n'],
            expected='from random import randint, choice\n\nn = randint(1, 3)\nprint(n)\n\n'
                     'for _ in range(n):\n    m, k = randint(1, 4), randint(2, 5)\n    print(m, k)\n'
        )

    @example
    def test_list_parameter_in_fixed(self, pattern):
        a = Integers(3)
        b = Integer()
        n = a[1]
        pattern.add(a)
        pattern.add(List(n, b))

        return dict(
            testcases=['\n2 1 4\n8\n', '3 2 5\n1\n4\n'],
            expected='from random import randint, choice\n\nn, m, k = randint(2, 3), randint(1, 2), randint(4, 5)\n'
                     'print(n, m, k)\n\nfor _ in range(m):\n    a = randint(1, 8)\n    print(a)\n'
        )

    @example
    def test_compound(self, pattern):
        t = Integer()
        n = Integer()
        l = List(n, Integer())
        f = List(t, Compound([n, l]))
        pattern.add(t)
        pattern.add(f)

        return dict(
            testcases=['\n2\n3\n1 2 3\n4\n1 2 3 4\n', '1\n5\n1 2 3 4 5\n'],
            expected='from random import randint, choice\n\nn = randint(1, 2)\nprint(n)\n\n'
                     'for _ in range(n):\n    m = randint(3, 5)\n    print(m)\n    for _ in range(m):\n'
                     '        a = randint(1, 5)\n        print(a)\n'
        )

    @example
    def test_string(self, pattern):
        pattern.add(String())

        return dict(
            testcases=['\nxyaz\n', 'verde\n'],
            expected='from random import randint, choice\n\n'
                     's = \'\'.join([choice("adervxyz") for _ in range(randint(4, 5))])\nprint(s)\n'
        )


class TestParser(unittest.TestCase):
    def test_parser_basics(self):
        p = Parser("hoy no es el dia!")
        self.assertEqual(p.peek(), 'h')
        self.assertEqual(p.token(), 'hoy')
        self.assertEqual(p.get(), ' ')
        self.assertEqual(p.get(), 'n')
        self.assertEqual(p.token(), 'o')

        with self.assertRaises(ParserError):
            p.token(allow_blank=False)

        for s in "es el dia!".split():
            self.assertEqual(p.token(), s)

        self.assertEqual(p.get(), None)

        with self.assertRaises(ParserError):
            p.token()

    def test_end(self):
        p = Parser("this is a token")
        p._pointer = 14
        token = p.token()
        self.assertEqual(token, "n")


if __name__ == '__main__':
    unittest.main()



from elements import List, Integer, Compound, Integers, String
from language import code

# Build pattern
pattern = code()
a = Integers(3)
b = Integer()
n = a[1]
pattern.add(a)
pattern.add(List(n, b))

prog = pattern.compile()

# Build testcases
# ['\n2 1 4\n8\n', '3 2 5\n1\n4\n']
testcases = """
2 1 4
8
-
3 2 5
1
4
"""

tc = testcases.split('-\n')

# Parse program
prog.parse(tc)

assert prog.ok

print(prog.generate())
exit(0)

# Print test
print("""@example
def TESTNAME(self, pattern):
    

    return dict(
        testcases={},
        expected={}
    )""".format(repr(tc), repr(prog.generate())))
# print(repr(prog.generate()))



from .elements import AbstractElement
from ..parser import Parser
from ..utils import rec_elements


class Compound(AbstractElement):
    def __init__(self, elems):
        self._elems = elems

    def parse(self, testcase: Parser):
        for e in self._elems:
            if not e.parse(testcase):
                return False
        return True

    def generate(self):
        return '\n'.join(e.generate() for e in self._elems)

    @rec_elements
    def elements(self):
        return self._elems



import argparse
from os import listdir
from os.path import join

from .generate import generate


def get_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('--version', action='store_true', default=False)
    parser.add_argument('-p', '--path', dest='path', default=None)
    parser.add_argument('-o', '--output', dest='output', default=None)
    parser.add_argument('-v', '--verbose', action='store_true')

    return parser.parse_args()


def get_testcases(args):
    if args.path is not None:
        testcases = []

        for fl in listdir(args.path):
            if fl.endswith('.in'):
                with open(join(args.path, fl)) as f:
                    testcases.append(f.read())

        if args.verbose:
            print("Loaded {} testcases.".format(len(testcases)))

        return testcases
    else:
        print("Provide testcases path using --path")
        exit(1)


def run(args):
    testcases = get_testcases(args)
    code = generate(testcases)

    if args.output is None:
        print(code)
    else:
        with open(args.output, 'w') as f:
            f.write(code)


if __name__ == '__main__':
    args = get_args()

    if args.version:
        from ._version import VERSION
        print(VERSION)
        exit(0)

    run(args)



from .elements import List, Integers, String, Integer
from .repository_meta import register_all
from ._pattern import PATTERNS


@register_all(head="1..=5")
def integers(head):
    yield Integers(head)


@register_all(head="1..=3", index="0..head", body="1..=5")
def list_head_index_body(head, index, body):
    h = Integers(head)
    yield h
    yield List(h[index], Integers(body))


@register_all(block0="1..4", block1="1..4")
def two_blocks_reversed_index(block0, block1):
    h = Integers(2)
    yield h
    yield List(h[1], Integers(block0))
    yield List(h[0], Integers(block1))


@register_all(block_int="1..4")
def two_blocks_string_integer(block_int):
    h = Integers(2)
    yield h
    yield List(h[0], String())
    yield List(h[1], Integers(block_int))


@register_all(head="1..=3", index="0..head")
def list_head_string(head, index):
    h = Integers(head)
    yield h
    yield List(h[index], String())


@register_all(head="1..3")
def board_square(head):
    h = Integers(head)
    yield h
    yield List(h[0], List(h[0], Integer()))


@register_all(head="2..=4")
def board_rectangle(head):
    h = Integers(head)
    yield h
    yield List(h[0], List(h[1], Integer()))



import setuptools
from tcgen._version import VERSION

with open("README.md", 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='tcgen',
    version=VERSION,
    author='Marcelo Fornet',
    author_email='mfornet94@gmail.com',
    description='Program synthesis tool. Read a list of strings and create a generator that sample strings from the same distribution as input strings. Focused on test cases of competitive programming problems.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='program synthesis competitive programming acm icpc codeforces',
    url='https://github.com/mfornet/tcgen',
    packages=setuptools.find_packages(),
    license='MIT',
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
)
